@@ -60,8 +60,7 @@ public class RemoveBlocks extends BlockTransferMessage {
 
   @Override
   public boolean equals(Object other) {
-    if (other instanceof RemoveBlocks) {
-      RemoveBlocks o = (RemoveBlocks) other;
+    if (other instanceof RemoveBlocks o) {
       return Objects.equals(appId, o.appId)
         && Objects.equals(execId, o.execId)
         && Arrays.equals(blockIds, o.blockIds);
