@@ -45,7 +45,7 @@ Options are:
   -n          : dry run mode. Performs checks and local builds, but do not upload anything.
   -t [tag]    : tag for the spark-rm docker image to use for building (default: "latest").
   -j [path]   : path to local JDK installation to use for building. By default the script will
-                use openjdk8 installed in the docker image.
+                use openjdk17 installed in the docker image.
   -s [step]   : runs a single step of the process; valid steps are: tag, build, docs, publish, finalize
 EOF
 }
@@ -72,11 +72,9 @@ fi
 
 if [ -d "$WORKDIR/output" ]; then
   if [ -z "$ANSWER" ]; then
-    read -p "Output directory already exists. Overwrite and continue? [y/n] " userinput
-    if [ "$userinput" != "y" ]; then
-      error "Exiting."
-    fi
-  elif [ "$ANSWER" != "y" ]; then
+    read -p "Output directory already exists. Overwrite and continue? [y/n] " ANSWER
+  fi
+  if [ "$ANSWER" != "y" ]; then
     error "Exiting."
   fi
 fi
@@ -84,11 +82,9 @@ fi
 if [ ! -z "$RELEASE_STEP" ] && [ "$RELEASE_STEP" = "finalize" ]; then
   echo "THIS STEP IS IRREVERSIBLE! Make sure the vote has passed and you pick the right RC to finalize."
   if [ -z "$ANSWER" ]; then
-    read -p "You must be a PMC member to run this step. Continue? [y/n] " userinput
-    if [ "$userinput" != "y" ]; then
-      error "Exiting."
-    fi
-  elif [ "$ANSWER" != "y" ]; then
+    read -p "You must be a PMC member to run this step. Continue? [y/n] " ANSWER
+  fi
+  if [ "$ANSWER" != "y" ]; then
     error "Exiting."
   fi
 
@@ -146,7 +142,6 @@ RELEASE_TAG=$RELEASE_TAG
 GIT_REF=$GIT_REF
 SPARK_PACKAGE_VERSION=$SPARK_PACKAGE_VERSION
 ASF_USERNAME=$ASF_USERNAME
-ASF_NEXUS_TOKEN=$ASF_NEXUS_TOKEN
 GIT_NAME=$GIT_NAME
 GIT_EMAIL=$GIT_EMAIL
 GPG_KEY=$GPG_KEY
@@ -156,9 +151,6 @@ GPG_PASSPHRASE=$GPG_PASSPHRASE
 RELEASE_STEP=$RELEASE_STEP
 USER=$USER
 DEBUG_MODE=$DEBUG_MODE
-ANSWER=$ANSWER
-GITHUB_ACTIONS=$GITHUB_ACTIONS
-SPARK_RC_COUNT=$SPARK_RC_COUNT
 EOF
 
 JAVA_VOL=
@@ -168,7 +160,7 @@ if [ -n "$JAVA" ]; then
 fi
 
 echo "Building $RELEASE_TAG; output will be at $WORKDIR/output"
-docker run $([ -z "$GITHUB_ACTIONS" ] && echo "-ti") \
+docker run -ti \
   --env-file "$ENVFILE" \
   --volume "$WORKDIR:/opt/spark-rm" \
   $JAVA_VOL \
