@@ -83,6 +83,7 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
   private static final Map<String, String> specialClasses = new HashMap<>();
   static {
     specialClasses.put("org.apache.spark.repl.Main", "spark-shell");
+    specialClasses.put("org.apache.spark.sql.application.ConnectRepl", "connect-shell");
     specialClasses.put("org.apache.spark.sql.hive.thriftserver.SparkSQLCLIDriver",
       SparkLauncher.NO_RESOURCE);
     specialClasses.put("org.apache.spark.sql.hive.thriftserver.HiveThriftServer2",
@@ -125,6 +126,7 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     boolean isExample = false;
     List<String> submitArgs = args;
     this.userArgs = Collections.emptyList();
+    isRemote |= "connect".equalsIgnoreCase(getApiMode(conf));
 
     if (args.size() > 0) {
       switch (args.get(0)) {
@@ -156,6 +158,17 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     }
   }
 
+  private static String getApiMode(Map<String, String> conf) {
+    boolean connectByDefault = "1".equals(System.getenv("SPARK_CONNECT_MODE"));
+    String defaultApiMode = connectByDefault ? "connect" : "classic";
+    String apiMode = conf.get(SparkLauncher.SPARK_API_MODE);
+    if ("classic".equalsIgnoreCase(apiMode) || "connect".equalsIgnoreCase(apiMode)) {
+      return apiMode;
+    } else {
+      return defaultApiMode;
+    }
+  }
+
   @Override
   public List<String> buildCommand(Map<String, String> env)
       throws IOException, IllegalArgumentException {
@@ -169,6 +182,10 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
   }
 
   List<String> buildSparkSubmitArgs() {
+    return buildSparkSubmitArgs(true);
+  }
+
+  List<String> buildSparkSubmitArgs(boolean includeRemote) {
     List<String> args = new ArrayList<>();
     OptionParser parser = new OptionParser(false);
     final boolean isSpecialCommand;
@@ -195,7 +212,7 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
       args.add(master);
     }
 
-    if (remote != null) {
+    if (includeRemote && remote != null) {
       args.add(parser.REMOTE);
       args.add(remote);
     }
@@ -211,8 +228,12 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     }
 
     for (Map.Entry<String, String> e : conf.entrySet()) {
-      args.add(parser.CONF);
-      args.add(String.format("%s=%s", e.getKey(), e.getValue()));
+      if (includeRemote ||
+           (!e.getKey().equalsIgnoreCase("spark.api.mode") &&
+             !e.getKey().equalsIgnoreCase("spark.remote"))) {
+        args.add(parser.CONF);
+        args.add(String.format("%s=%s", e.getKey(), e.getValue()));
+      }
     }
 
     if (propertiesFile != null) {
@@ -245,13 +266,21 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
 
     if (mainClass != null) {
       args.add(parser.CLASS);
-      args.add(mainClass);
+      if (isRemote && "1".equals(getenv("SPARK_SCALA_SHELL"))) {
+        args.add("org.apache.spark.sql.application.ConnectRepl");
+      } else {
+        args.add(mainClass);
+      }
     }
 
     args.addAll(parsedArgs);
 
     if (appResource != null) {
-      args.add(appResource);
+      if (isRemote && "1".equals(getenv("SPARK_SCALA_SHELL"))) {
+        args.add("connect-shell");
+      } else {
+        args.add(appResource);
+      }
     }
 
     args.addAll(appArgs);
@@ -267,6 +296,12 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     Map<String, String> config = getEffectiveConfig();
     boolean isClientMode = isClientMode(config);
     String extraClassPath = isClientMode ? config.get(SparkLauncher.DRIVER_EXTRA_CLASSPATH) : null;
+    String defaultExtraClassPath = config.get(SparkLauncher.DRIVER_DEFAULT_EXTRA_CLASS_PATH);
+    if (extraClassPath == null || extraClassPath.trim().isEmpty()) {
+      extraClassPath = defaultExtraClassPath;
+    } else {
+      extraClassPath += File.pathSeparator + defaultExtraClassPath;
+    }
 
     List<String> cmd = buildJavaCommand(extraClassPath);
     // Take Thrift/Connect Server as daemon
@@ -303,19 +338,24 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
         config.get(SparkLauncher.DRIVER_EXTRA_LIBRARY_PATH));
     }
 
-    // SPARK-36796: Always add default `--add-opens` to submit command
+    // SPARK-36796: Always add some JVM runtime default options to submit command
     addOptionString(cmd, JavaModuleOptions.defaultModuleOptions());
+    addOptionString(cmd, "-Dderby.connection.requireAuthentication=false");
     cmd.add("org.apache.spark.deploy.SparkSubmit");
     cmd.addAll(buildSparkSubmitArgs());
     return cmd;
   }
 
   private void checkJavaOptions(String javaOptions) {
-    if (!isEmpty(javaOptions) && javaOptions.contains("Xmx")) {
-      String msg = String.format("Not allowed to specify max heap(Xmx) memory settings through " +
-        "java options (was %s). Use the corresponding --driver-memory or " +
-        "spark.driver.memory configuration instead.", javaOptions);
-      throw new IllegalArgumentException(msg);
+    if (!isEmpty(javaOptions)) {
+      for (String javaOption: CommandBuilderUtils.parseOptionString(javaOptions)) {
+        if (javaOption.startsWith("-Xmx")) {
+            String msg = String.format("Not allowed to specify max heap(Xmx) memory settings " +
+              "through java options (was %s). Use the corresponding --driver-memory or " +
+              "spark.driver.memory configuration instead.", javaOptions);
+            throw new IllegalArgumentException(msg);
+        }
+      }
     }
   }
 
@@ -334,7 +374,8 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     // When launching the pyspark shell, the spark-submit arguments should be stored in the
     // PYSPARK_SUBMIT_ARGS env variable.
     appResource = PYSPARK_SHELL_RESOURCE;
-    constructEnvVarArgs(env, "PYSPARK_SUBMIT_ARGS");
+    // Do not pass remote configurations to Spark Connect server via Py4J.
+    constructEnvVarArgs(env, "PYSPARK_SUBMIT_ARGS", false);
 
     // Will pick up the binary executable in the following order
     // 1. conf spark.pyspark.driver.python
@@ -357,13 +398,18 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     String masterStr = firstNonEmpty(master, conf.getOrDefault(SparkLauncher.SPARK_MASTER, null));
     String deployStr = firstNonEmpty(
       deployMode, conf.getOrDefault(SparkLauncher.DEPLOY_MODE, null));
-    if (!conf.containsKey(SparkLauncher.SPARK_LOCAL_REMOTE) &&
-        remoteStr != null && (masterStr != null || deployStr != null)) {
+    if (remoteStr != null && (masterStr != null || deployStr != null)) {
       throw new IllegalStateException("Remote cannot be specified with master and/or deploy mode.");
     }
+
+    String apiMode = getApiMode(conf);
+    env.put("SPARK_API_MODE", apiMode);
     if (remoteStr != null) {
       env.put("SPARK_REMOTE", remoteStr);
       env.put("SPARK_CONNECT_MODE_ENABLED", "1");
+    } else if ("connect".equalsIgnoreCase(apiMode)) {
+      env.put("MASTER", firstNonEmpty(masterStr, "local"));
+      env.put("SPARK_CONNECT_MODE_ENABLED", "1");
     }
 
     if (!isEmpty(pyOpts)) {
@@ -383,7 +429,7 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     // When launching the SparkR shell, store the spark-submit arguments in the SPARKR_SUBMIT_ARGS
     // env variable.
     appResource = SPARKR_SHELL_RESOURCE;
-    constructEnvVarArgs(env, "SPARKR_SUBMIT_ARGS");
+    constructEnvVarArgs(env, "SPARKR_SUBMIT_ARGS", true);
 
     // Set shell.R as R_PROFILE_USER to load the SparkR package when the shell comes up.
     String sparkHome = System.getenv("SPARK_HOME");
@@ -398,12 +444,13 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
 
   private void constructEnvVarArgs(
       Map<String, String> env,
-      String submitArgsEnvVariable) throws IOException {
+      String submitArgsEnvVariable,
+      boolean includeRemote) throws IOException {
     mergeEnvPathList(env, getLibPathEnvName(),
       getEffectiveConfig().get(SparkLauncher.DRIVER_EXTRA_LIBRARY_PATH));
 
     StringBuilder submitArgs = new StringBuilder();
-    for (String arg : buildSparkSubmitArgs()) {
+    for (String arg : buildSparkSubmitArgs(includeRemote)) {
       if (submitArgs.length() > 0) {
         submitArgs.append(" ");
       }
@@ -486,37 +533,34 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
     @Override
     protected boolean handle(String opt, String value) {
       switch (opt) {
-        case MASTER:
-          master = value;
-          break;
-        case REMOTE:
+        case MASTER -> master = value;
+        case REMOTE -> {
+          isRemote = true;
           remote = value;
-          break;
-        case DEPLOY_MODE:
-          deployMode = value;
-          break;
-        case PROPERTIES_FILE:
-          propertiesFile = value;
-          break;
-        case DRIVER_MEMORY:
-          conf.put(SparkLauncher.DRIVER_MEMORY, value);
-          break;
-        case DRIVER_JAVA_OPTIONS:
-          conf.put(SparkLauncher.DRIVER_EXTRA_JAVA_OPTIONS, value);
-          break;
-        case DRIVER_LIBRARY_PATH:
-          conf.put(SparkLauncher.DRIVER_EXTRA_LIBRARY_PATH, value);
-          break;
-        case DRIVER_CLASS_PATH:
-          conf.put(SparkLauncher.DRIVER_EXTRA_CLASSPATH, value);
-          break;
-        case CONF:
+        }
+        case DEPLOY_MODE -> deployMode = value;
+        case PROPERTIES_FILE -> propertiesFile = value;
+        case DRIVER_MEMORY -> conf.put(SparkLauncher.DRIVER_MEMORY, value);
+        case DRIVER_JAVA_OPTIONS -> conf.put(SparkLauncher.DRIVER_EXTRA_JAVA_OPTIONS, value);
+        case DRIVER_LIBRARY_PATH -> conf.put(SparkLauncher.DRIVER_EXTRA_LIBRARY_PATH, value);
+        case DRIVER_DEFAULT_CLASS_PATH ->
+          conf.put(SparkLauncher.DRIVER_DEFAULT_EXTRA_CLASS_PATH, value);
+        case DRIVER_CLASS_PATH -> conf.put(SparkLauncher.DRIVER_EXTRA_CLASSPATH, value);
+        case CONF -> {
           checkArgument(value != null, "Missing argument to %s", CONF);
           String[] setConf = value.split("=", 2);
           checkArgument(setConf.length == 2, "Invalid argument to %s: %s", CONF, value);
+          // If both spark.remote and spark.mater are set, the error will be thrown later when
+          // the application is started.
+          if (setConf[0].equals("spark.remote")) {
+            isRemote = true;
+          } else if (setConf[0].equals(SparkLauncher.SPARK_API_MODE)) {
+            // Respects if the API mode is explicitly set.
+            isRemote = setConf[1].equalsIgnoreCase("connect");
+          }
           conf.put(setConf[0], setConf[1]);
-          break;
-        case CLASS:
+        }
+        case CLASS -> {
           // The special classes require some special command line handling, since they allow
           // mixing spark-submit arguments with arguments that should be propagated to the shell
           // itself. Note that for this to work, the "--class" argument must come before any
@@ -526,25 +570,22 @@ class SparkSubmitCommandBuilder extends AbstractCommandBuilder {
             allowsMixedArguments = true;
             appResource = specialClasses.get(value);
           }
-          break;
-        case KILL_SUBMISSION:
-        case STATUS:
+        }
+        case KILL_SUBMISSION, STATUS -> {
           isSpecialCommand = true;
           parsedArgs.add(opt);
           parsedArgs.add(value);
-          break;
-        case HELP:
-        case USAGE_ERROR:
-        case VERSION:
+        }
+        case HELP, USAGE_ERROR, VERSION -> {
           isSpecialCommand = true;
           parsedArgs.add(opt);
-          break;
-        default:
+        }
+        default -> {
           parsedArgs.add(opt);
           if (value != null) {
             parsedArgs.add(value);
           }
-          break;
+        }
       }
       return true;
     }
