@@ -16,19 +16,26 @@
  */
 package org.apache.spark.sql.execution
 
+import scala.collection.mutable
 import scala.io.Source
+import scala.util.Try
 
-import org.apache.spark.sql.{AnalysisException, Dataset, FastOperator}
-import org.apache.spark.sql.catalyst.{QueryPlanningTracker, QueryPlanningTrackerCallback}
-import org.apache.spark.sql.catalyst.analysis.UnresolvedNamespace
-import org.apache.spark.sql.catalyst.expressions.UnsafeRow
+import org.apache.spark.sql.{AnalysisException, ExtendedExplainGenerator, FastOperator}
+import org.apache.spark.sql.catalyst.{QueryPlanningTracker, QueryPlanningTrackerCallback, TableIdentifier}
+import org.apache.spark.sql.catalyst.analysis.{CurrentNamespace, UnresolvedFunction, UnresolvedRelation}
+import org.apache.spark.sql.catalyst.expressions.{Alias, UnsafeRow}
 import org.apache.spark.sql.catalyst.plans.QueryPlan
 import org.apache.spark.sql.catalyst.plans.logical.{CommandResult, LogicalPlan, OneRowRelation, Project, ShowTables, SubqueryAlias}
 import org.apache.spark.sql.catalyst.trees.TreeNodeTag
+import org.apache.spark.sql.catalyst.util.StringUtils.PlanStringConcat
+import org.apache.spark.sql.classic.Dataset
 import org.apache.spark.sql.connector.catalog.CatalogManager.SESSION_CATALOG_NAME
+import org.apache.spark.sql.execution.adaptive.{AdaptiveSparkPlanExec, QueryStageExec}
 import org.apache.spark.sql.execution.datasources.v2.ShowTablesExec
+import org.apache.spark.sql.execution.joins.SortMergeJoinExec
 import org.apache.spark.sql.internal.SQLConf
 import org.apache.spark.sql.test.SharedSparkSession
+import org.apache.spark.storage.ShuffleIndexBlockId
 import org.apache.spark.util.Utils
 
 case class QueryExecutionTestRecord(
@@ -44,7 +51,7 @@ class QueryExecutionSuite extends SharedSparkSession {
 
   def checkDumpedPlans(path: String, expected: Int): Unit = Utils.tryWithResource(
     Source.fromFile(path)) { source =>
-    assert(source.getLines.toList
+    assert(source.getLines().toList
       .takeWhile(_ != "== Whole Stage Codegen ==") == List(
       "== Parsed Logical Plan ==",
       s"Range (0, $expected, step=1, splits=Some(2))",
@@ -160,7 +167,9 @@ class QueryExecutionSuite extends SharedSparkSession {
 
     // Throw an AnalysisException - this should be captured.
     spark.experimental.extraStrategies = Seq[SparkStrategy](
-      (_: LogicalPlan) => throw new AnalysisException("exception"))
+      (_: LogicalPlan) => throw new AnalysisException(
+        "UNSUPPORTED_DATASOURCE_FOR_DIRECT_QUERY",
+        messageParameters = Map("dataSourceType" -> "XXX")))
     assert(qe.toString.contains("org.apache.spark.sql.AnalysisException"))
 
     // Throw an Error - this should not be captured.
@@ -250,22 +259,22 @@ class QueryExecutionSuite extends SharedSparkSession {
     def qe(logicalPlan: LogicalPlan, callback: QueryPlanningTrackerCallback): QueryExecution =
       new QueryExecution(spark, logicalPlan, new QueryPlanningTracker(Some(callback)))
 
-    val showTables = ShowTables(UnresolvedNamespace(Seq.empty[String]), None)
+    val showTables = ShowTables(CurrentNamespace, None)
     val showTablesQe = qe(showTables, mockCallback1)
-    showTablesQe.assertAnalyzed
-    mockCallback1.assertAnalyzed
+    showTablesQe.assertAnalyzed()
+    mockCallback1.assertAnalyzed()
     assert(showTablesQe.commandExecuted.isInstanceOf[CommandResult])
-    mockCallback1.assertCommandExecuted
+    mockCallback1.assertCommandExecuted()
     assert(showTablesQe.executedPlan.isInstanceOf[CommandResultExec])
     val showTablesResultExec = showTablesQe.executedPlan.asInstanceOf[CommandResultExec]
     assert(showTablesResultExec.commandPhysicalPlan.isInstanceOf[ShowTablesExec])
 
     val project = Project(showTables.output, SubqueryAlias("s", showTables))
     val projectQe = qe(project, mockCallback2)
-    projectQe.assertAnalyzed
-    mockCallback2.assertAnalyzed
+    projectQe.assertAnalyzed()
+    mockCallback2.assertAnalyzed()
     assert(projectQe.commandExecuted.isInstanceOf[Project])
-    mockCallback2.assertCommandExecuted
+    mockCallback2.assertCommandExecuted()
     assert(projectQe.commandExecuted.children.length == 1)
     assert(projectQe.commandExecuted.children(0).isInstanceOf[SubqueryAlias])
     assert(projectQe.commandExecuted.children(0).children.length == 1)
@@ -278,34 +287,76 @@ class QueryExecutionSuite extends SharedSparkSession {
   test("SPARK-44145: non eagerly executed command setReadyForExecution") {
     val mockCallback = MockCallback()
 
-    val showTables = ShowTables(UnresolvedNamespace(Seq.empty[String]), None)
+    val showTables = ShowTables(CurrentNamespace, None)
     val showTablesQe = new QueryExecution(
       spark,
       showTables,
       new QueryPlanningTracker(Some(mockCallback)),
       CommandExecutionMode.SKIP)
-    showTablesQe.assertAnalyzed
-    mockCallback.assertAnalyzed
-    showTablesQe.assertOptimized
-    mockCallback.assertOptimized
-    showTablesQe.assertSparkPlanPrepared
-    mockCallback.assertSparkPlanPrepared
-    showTablesQe.assertExecutedPlanPrepared
-    mockCallback.assertExecutedPlanPrepared
+    showTablesQe.assertAnalyzed()
+    mockCallback.assertAnalyzed()
+    showTablesQe.assertOptimized()
+    mockCallback.assertOptimized()
+    showTablesQe.assertSparkPlanPrepared()
+    mockCallback.assertSparkPlanPrepared()
+    showTablesQe.assertExecutedPlanPrepared()
+    mockCallback.assertExecutedPlanPrepared()
   }
 
   test("SPARK-44145: Plan setReadyForExecution") {
     val mockCallback = MockCallback()
     val plan: LogicalPlan = org.apache.spark.sql.catalyst.plans.logical.Range(0, 1, 1, 1)
     val df = Dataset.ofRows(spark, plan, new QueryPlanningTracker(Some(mockCallback)))
-    df.queryExecution.assertAnalyzed
-    mockCallback.assertAnalyzed
-    df.queryExecution.assertOptimized
-    mockCallback.assertOptimized
-    df.queryExecution.assertSparkPlanPrepared
-    mockCallback.assertSparkPlanPrepared
-    df.queryExecution.assertExecutedPlanPrepared
-    mockCallback.assertExecutedPlanPrepared
+    df.queryExecution.assertAnalyzed()
+    mockCallback.assertAnalyzed()
+    df.queryExecution.assertOptimized()
+    mockCallback.assertOptimized()
+    df.queryExecution.assertSparkPlanPrepared()
+    mockCallback.assertSparkPlanPrepared()
+    df.queryExecution.assertExecutedPlanPrepared()
+    mockCallback.assertExecutedPlanPrepared()
+  }
+
+  private def cleanupShuffles(): Unit = {
+    val blockManager = spark.sparkContext.env.blockManager
+    blockManager.diskBlockManager.getAllBlocks().foreach {
+      case ShuffleIndexBlockId(shuffleId, _, _) =>
+        spark.sparkContext.shuffleDriverComponents.removeShuffle(shuffleId, true)
+      case _ =>
+    }
+  }
+
+  test("SPARK-47764: Cleanup shuffle dependencies - DoNotCleanup mode") {
+    val plan = spark.range(100).repartition(10).logicalPlan
+    val df = Dataset.ofRows(spark, plan, DoNotCleanup)
+    df.collect()
+
+    val blockManager = spark.sparkContext.env.blockManager
+    assert(blockManager.migratableResolver.getStoredShuffles().nonEmpty)
+    assert(blockManager.diskBlockManager.getAllBlocks().nonEmpty)
+    cleanupShuffles()
+  }
+
+  test("SPARK-47764: Cleanup shuffle dependencies - SkipMigration mode") {
+    val plan = spark.range(100).repartition(10).logicalPlan
+    val df = Dataset.ofRows(spark, plan, SkipMigration)
+    df.collect()
+
+    val blockManager = spark.sparkContext.env.blockManager
+    assert(blockManager.migratableResolver.getStoredShuffles().isEmpty)
+    assert(blockManager.diskBlockManager.getAllBlocks().nonEmpty)
+    cleanupShuffles()
+  }
+
+  test("SPARK-47764: Cleanup shuffle dependencies - RemoveShuffleFiles mode") {
+    val plan = spark.range(100).repartition(10).logicalPlan
+    val df = Dataset.ofRows(spark, plan, RemoveShuffleFiles)
+    df.collect()
+
+    val blockManager = spark.sparkContext.env.blockManager
+    assert(blockManager.migratableResolver.getStoredShuffles().isEmpty)
+    assert(blockManager.diskBlockManager.getAllBlocks().isEmpty)
+    cleanupShuffles()
   }
 
   test("SPARK-35378: Return UnsafeRow in CommandResultExecCheck execute methods") {
@@ -334,6 +385,60 @@ class QueryExecutionSuite extends SharedSparkSession {
     }
   }
 
+  test("SPARK-47289: extended explain info") {
+    val concat = new PlanStringConcat()
+    withSQLConf(SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "true",
+      SQLConf.AUTO_BROADCASTJOIN_THRESHOLD.key -> "-1",
+      SQLConf.EXTENDED_EXPLAIN_PROVIDERS.key -> "org.apache.spark.sql.execution.ExtendedInfo") {
+      val left = Seq((1, 10L), (2, 100L), (3, 1000L)).toDF("l1", "l2")
+      val right = Seq((1, 2), (2, 3), (3, 5)).toDF("r1", "r2")
+      val data = left.join(right, $"l1" === $"r1")
+      val df = data.selectExpr("l2 + r2")
+      // execute the plan so that the final adaptive plan is available
+      df.collect()
+      val qe = df.queryExecution
+      qe.extendedExplainInfo(concat.append, qe.executedPlan)
+      val info = concat.toString
+      val expected = "\n== Extended Information (Test) ==\n" +
+        "Scan Info: LocalTableScan\n" +
+        "Project Info: Project\n" +
+        "SMJ Info: SortMergeJoin"
+      assert(info == expected)
+    }
+  }
+
+  test("SPARK-50600: Failed analysis should send analyzed event") {
+    val mockCallback = MockCallback()
+
+    def table(ref: String): LogicalPlan = UnresolvedRelation(TableIdentifier(ref))
+
+    val unresolvedUndefinedFunc = UnresolvedFunction("unknown", Seq.empty, isDistinct = false)
+    val plan = Project(Seq(Alias(unresolvedUndefinedFunc, "call1")()), table("table"))
+    val dataset = Try {
+      val df = Dataset.ofRows(spark, plan, new QueryPlanningTracker(Some(mockCallback)))
+      df.queryExecution.assertAnalyzed()
+    }
+    assert(dataset.failed.get.isInstanceOf[AnalysisException])
+    mockCallback.assertAnalyzed()
+  }
+
+  test("SPARK-51265: IncrementalExecution should set the command execution code correctly") {
+    withTempView("s") {
+      val streamDf = spark.readStream.format("rate").load()
+      streamDf.createOrReplaceTempView("s")
+      withTable("output") {
+        val ex = intercept[AnalysisException] {
+          // Creates a table from streaming source with batch query. This should fail.
+          spark.sql("CREATE TABLE output USING csv AS SELECT * FROM s")
+        }
+        assert(
+          ex.getMessage.contains("Queries with streaming sources must be executed with " +
+            "writeStream.start()")
+        )
+      }
+    }
+  }
+
   case class MockCallbackEagerCommand(
       var trackerAnalyzed: QueryPlanningTracker = null,
       var trackerReadyForExecution: QueryPlanningTracker = null)
@@ -376,6 +481,15 @@ class QueryExecutionSuite extends SharedSparkSession {
       var trackerAnalyzed: QueryPlanningTracker = null,
       var trackerReadyForExecution: QueryPlanningTracker = null)
       extends QueryPlanningTrackerCallback {
+    override def analysisFailed(
+        trackerFromCallback: QueryPlanningTracker,
+        analyzedPlan: LogicalPlan): Unit = {
+      trackerAnalyzed = trackerFromCallback
+      assert(!trackerAnalyzed.phases.keySet.contains(QueryPlanningTracker.ANALYSIS))
+      assert(!trackerAnalyzed.phases.keySet.contains(QueryPlanningTracker.OPTIMIZATION))
+      assert(!trackerAnalyzed.phases.keySet.contains(QueryPlanningTracker.PLANNING))
+      assert(analyzedPlan != null)
+    }
     def analyzed(trackerFromCallback: QueryPlanningTracker, plan: LogicalPlan): Unit = {
       trackerAnalyzed = trackerFromCallback
       assert(trackerAnalyzed.phases.keySet.contains(QueryPlanningTracker.ANALYSIS))
@@ -411,3 +525,45 @@ class QueryExecutionSuite extends SharedSparkSession {
     }
   }
 }
+
+class ExtendedInfo extends ExtendedExplainGenerator {
+
+  override def title: String = "Test"
+
+  override def generateExtendedInfo(plan: SparkPlan): String = {
+    val info = mutable.LinkedHashSet[String]() // don't allow duplicates
+    extensionInfo(plan, info)
+    info.mkString("\n")
+  }
+
+  def getActualPlan(plan: SparkPlan): SparkPlan = {
+    plan match {
+      case p : AdaptiveSparkPlanExec => getActualPlan(p.executedPlan)
+      case p : QueryStageExec => p.plan
+      case p : WholeStageCodegenExec => p.child
+      case p => p
+    }
+  }
+
+  def extensionInfo(p: SparkPlan, info: mutable.Set[String]): Unit = {
+    //    val info = mutable.Set[String]() // don't allow duplicates
+    val actualPlan = getActualPlan(p)
+    if (actualPlan.innerChildren.nonEmpty) {
+      actualPlan.innerChildren.foreach(c => {
+        extensionInfo(c.asInstanceOf[SparkPlan], info)
+      })
+    }
+    if (actualPlan.children.nonEmpty) {
+      actualPlan.children.foreach(c => {
+        extensionInfo(c, info)
+      })
+    }
+    actualPlan match {
+      case p : LocalTableScanExec => info += s"Scan Info: ${p.nodeName}"
+      case p : SortMergeJoinExec => info += s"SMJ Info: ${p.nodeName}"
+      case p : ProjectExec => info += s"Project Info: ${p.nodeName}"
+      case _ =>
+    }
+    ()
+  }
+}
